from ..engine import Engine
from ..query.builder import Builder


class Zenatonable:
    """
        Sends self as the single job to be executed
        to the engine and returns the result
    """
    def dispatch(self):
        return Engine().dispatch([self])

    """
        Sends self as the single job to be dispatched
        to the engine and returns the result
    """
    def execute(self):
        return Engine().execute([self])[0]

    """
        Search for workflows to interact with.
        For available methods, see .query.builder.Builder
        :param String id ID for a given worflow
        :returns .query.builder.Builder a query builder object
    """

    @classmethod
    def where_id(cls, workflow_id):
        return Builder(cls).where_id(workflow_id)



from .abstracts.task import Task
from .abstracts.workflow import Workflow
from .exceptions import InvalidArgumentError
from .singleton import Singleton


class Engine(metaclass=Singleton):
    """
        Zenaton Engine is a singleton class that stores a reference to the current
        client and processor. It then handles job processing either locally or
        through the processor with Zenaton workers
        To access the instance, call `Zenaton::Engine.instance`
    """

    def __init__(self):
        from .client import Client
        self.client = Client()
        self.processor = None

    """
        Executes jobs synchronously
        @param jobs [Array<Zenaton::Interfaces::Job>]
        @return [Array<String>, nil] the results if executed locally, or nil
    """
    def execute(self, jobs):
        for job in jobs:
            Engine._check_argument(job)
        if len(jobs) == 0 or self.processor is None:
            return [job.handle() for job in jobs]
        return self.processor.process(jobs, True)

    """
        Executes jobs asynchronously
        @param jobs [Array<Zenaton::Interfaces::Job>]
        @return nil
    """
    def dispatch(self, jobs):
        map(Engine._check_argument, jobs)
        if len(jobs) == 0 or self.processor is None:
            [self.local_dispatch(job) for job in jobs]
        if self.processor and len(jobs) > 0:
            self.processor.process(jobs, False)

    def local_dispatch(self, job):
        if isinstance(job, Workflow):
            self.client.start_workflow(job)
        else:
            self.client.start_task(job)

    @staticmethod
    def _check_argument(job):
        if not Engine.valid_job(job):
            raise InvalidArgumentError('You can only execute or dispatch Zenaton Task or Worflow')

    """
        Checks if the job is a valid job i.e. it is either a Task or a Workflow
    """
    @staticmethod
    def valid_job(job):
        return isinstance(job, (Task, Workflow))



import os

import pytest
from dotenv import load_dotenv

from zenaton.client import Client

# LOADING CONFIG FROM .env file
load_dotenv()
app_id = os.getenv('ZENATON_APP_ID')
api_token = os.getenv('ZENATON_API_TOKEN')
app_env = os.getenv('ZENATON_APP_ENV')


@pytest.fixture
def client():
    return Client(app_id, api_token, app_env)



"""Zenaton base error class"""
class Error(Exception):
    pass


"""Exception raised when communication with workers failed"""
class InternalError(Exception):
    pass


"""Exception raise when clien code is invalid"""
class ExternalError(Exception):
    pass


"""Exception raised when wrong argument type is provided"""
class InvalidArgumentError(ExternalError):
    pass


"""Exception raised when the workflow is unknown"""
class UnknownWorkflowError(ExternalError):
    pass



import copy
import datetime
import pytest

from freezegun import freeze_time
import pytz

from zenaton.exceptions import ExternalError
from zenaton.tasks.wait import Wait
from zenaton.traits.with_timestamp import WithTimestamp


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_overflow_pos(wait):
    now = datetime.datetime.now()
    with pytest.raises(ValueError):
        wait.day_of_month(32)
        wait.get_timetamp_or_duration()


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_overflow_zero(wait):
    now = datetime.datetime.now()
    with pytest.raises(ValueError):
        wait.day_of_month(0)
        wait.get_timetamp_or_duration()


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_overflow_neg(wait):
    now = datetime.datetime.now()
    with pytest.raises(ValueError):
        wait.day_of_month(-1)
        wait.get_timetamp_or_duration()


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_end_of_month_no_skip(wait):
    wait.day_of_month(28)
    assert wait.get_timetamp_or_duration()[0] == 1488283200  # i.e: "2017-02-28 12:00:00"


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_end_of_month_next(wait):
    wait.day_of_month(1)
    assert wait.get_timetamp_or_duration()[0] == 1488369600  # i.e: "2017-03-01 12:00:00"


@freeze_time("2017-02-03 12:00:00")
@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_end_of_month_skip(wait):
    wait.day_of_month(31)
    assert wait.get_timetamp_or_duration()[0] == 1490961600 # i.e: "2017-03-31 12:00:00"


@pytest.mark.usefixtures("wait")
@freeze_time("2017-02-03 12:01:02")  # It's a Friday.
def test_get_timetamp_or_duration_next_weekday(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now(pytz.timezone('Europe/Paris'))
    now_day = (now.weekday() + 1) % 7
    weekday = WithTimestamp.WEEKDAYS[now.weekday()]
    getattr(wait2, weekday)()
    assert wait2.get_timetamp_or_duration()[0] - int(now.timestamp()) == (7 * 24 * 60 * 60)
    for i in range(1, 6):
        new_wait = copy.deepcopy(wait)
        getattr(new_wait, new_wait.WEEKDAYS[now_day])(i)
        assert new_wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == (24 * 60 * 60) * (1 + (7 * (i - 1)))


@pytest.mark.usefixtures("wait")
@freeze_time("2017-02-03 11:01:02")  # It's a Friday, and 11:01 UTC is 12:01 in Paris that day.
def test_get_timetamp_or_duration_weekday(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now(pytz.timezone('Europe/Paris'))
    wait.friday().at("13:01:02")
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == (60 * 60)
    wait2.friday().at("11:01:02")
    assert wait2.get_timetamp_or_duration()[0] - int(now.timestamp()) == (7 * 24 * 60 * 60) - (60 * 60)


@pytest.mark.usefixtures("wait")
@freeze_time("2017-02-03 11:01:02")
def test_get_timetamp_or_duration_day_of_month(wait):
    wait2 = copy.deepcopy(wait)
    wait3 = copy.deepcopy(wait)
    now = datetime.datetime.now(pytz.timezone('Europe/Paris'))
    wait.day_of_month(now.day + 1)
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 24 * 60 * 60
    wait2.day_of_month(now.day - 1)
    wait_duration = wait2.get_timetamp_or_duration()[0] - int(now.timestamp())
    assert 27 * 24 * 60 * 60 <= wait_duration <= 30 * 24 * 60 * 60
    wait3.day_of_month(now.day)
    wait_duration = wait3.get_timetamp_or_duration()[0] - int(now.timestamp())
    assert 28 * 24 * 60 * 60 <= wait_duration <= 31 * 24 * 60 * 60


@pytest.mark.usefixtures("wait")
@freeze_time("2017-02-03 11:01:02")  # 11:01 UTC is actually 12:01 Paris time that day.
def test_get_timetamp_or_duration_day_of_month_at(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now(pytz.timezone('Europe/Paris'))
    wait.day_of_month(3).at("13:01:02")
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 60 * 60
    wait2.day_of_month(3).at("11:01:02")
    assert 28 * 24 * 60 * 60 - 60 * 60 == wait2.get_timetamp_or_duration()[0] - int(now.timestamp())


@pytest.mark.usefixtures("wait")
@freeze_time("2017-02-03 11:01:02")  # 11:01 UTC is actually 12:01 Paris time that day.
def test_get_timetamp_or_duration_at(wait):
    now = datetime.datetime.now(pytz.timezone('Europe/Paris'))
    wait.set_timezone('Europe/Paris')
    wait.at('13:01:02')
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 60 * 60


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_timestamp(wait):
    now_timestamp = int(datetime.datetime.now().timestamp())
    wait.timestamp(now_timestamp + 10.0)
    assert wait.get_timetamp_or_duration()[0] - now_timestamp == 10


@pytest.mark.usefixtures("wait")
def test_get_timetamp_push(wait):
    for name in ['timestamp', 'at', 'day_of_month', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday',
                 'sunday']:
        getattr(wait, name)(1)
        assert wait.buffer.get(name, None)


@pytest.mark.usefixtures("wait")
def test_set_mode(wait):
    wait._WithTimestamp__set_mode(wait.MODE_WEEK_DAY)
    assert wait.mode == wait.MODE_WEEK_DAY
    with pytest.raises(ExternalError):
        wait._WithTimestamp__set_mode(wait.MODE_WEEK_DAY)
    wait._WithTimestamp__set_mode(wait.MODE_MONTH_DAY)
    assert wait.mode == wait.MODE_AT


@pytest.mark.usefixtures("wait")
def test_is_timestamp_mode_set(wait):
    wait.mode = wait.MODE_WEEK_DAY
    assert not wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_WEEK_DAY)
    assert wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_TIMESTAMP)
    assert not wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_WEEK_DAY)


@pytest.mark.usefixtures("wait")
def test_set_timezone(wait):
    wait.set_timezone('Africa/Libreville')
    assert wait.timezone == 'Africa/Libreville'
    new_wait = Wait()
    assert new_wait.timezone == 'Africa/Libreville'
    with pytest.raises(ExternalError):
        wait.set_timezone('Africa/Leconi')


@pytest.mark.usefixtures("wait")
def test_valid_timezone(wait):
    assert wait._WithTimestamp__is_valid_timezone('Africa/Libreville')
    assert not wait._WithTimestamp__is_valid_timezone('Africa/Mvengué')



from .engine import Engine


class Parallel:
    """
        Build a collection of jobs to be executed in parallel
        :params: [.abstracts.job.Job] items
    """

    def __init__(self, *items):
        self.items = items

    # Execute synchronous jobs
    def execute(self):
        return Engine().execute(self.items)

    # Dispatches synchronous jobs
    def dispatch(self):
        return Engine().dispatch(self.items)



import os
import pytest
import pytest_mock

from zenaton.exceptions import InvalidArgumentError


class FakeTask:
    pass


@pytest.mark.usefixtures("engine")
def test_check_argument(engine):
    with pytest.raises(InvalidArgumentError):
        engine.execute([FakeTask()])


@pytest.mark.usefixtures("engine", "task0", "sequential_workflow")
def test_valid_job(engine, sequential_workflow, task0):
    assert not engine.valid_job(FakeTask())
    assert engine.valid_job(sequential_workflow)
    assert engine.valid_job(task0)


@pytest.mark.usefixtures("client", "engine", "task0")
@pytest.mark.skipif(not os.getenv('ZENATON_API_TOKEN'), reason="requires an API token")
def test_dispatch_task(client, engine, task0, mocker):
    mocker.spy(client, "start_task")
    task0.dispatch()
    assert client.start_task.call_count == 1



"""
Zenaton client library
"""
from __future__ import absolute_import, print_function, unicode_literals

import os
from io import open

from setuptools import find_packages, setup

import zenaton


def read(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, encoding='utf-8') as f:
        return f.read()

TEST_DEPS = [
    'freezegun',
    'pytest',
    'pytest_mock',
    'python-dotenv',
]

setup(
    name='zenaton',
    version=zenaton.__version__,
    author='Zenaton',
    author_email='yann@zenaton.com',
    description='Zenaton client library',
    long_description=read('README.md') + '\n' + read('CHANGELOG.md'),
    long_description_content_type='text/markdown',
    keywords='workflow tasks queue orchestration scheduling',
    url='https://zenaton.com/',
    license='Apache License, Version 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3',
    install_requires=[
        'requests',
        'pytz'
    ],
    tests_require=TEST_DEPS,
    extras_require={'test': TEST_DEPS},
    zip_safe=False
)



import abc

from ..abstracts.workflow import Workflow
from ..exceptions import ExternalError
from ..traits.zenatonable import Zenatonable


class Version(Workflow, Zenatonable):

    def __init__(self, *args):
        if args:
            self.args = args

    @abc.abstractmethod
    def versions(self):
        raise NotImplementedError("Please override the `versions' method in your subclass")

    """Calls handle on the current implementation"""

    def handle(self):
        self.current_implementation().handle()

    """
    Get the current implementation class
    returns class
    """

    def current(self):
        return self.__get_versions()[-1]

    """
        Get the first implementation class
        returns class
    """

    def initial(self):
        return self.__get_versions()[0]

    """
    Returns an instance of the current implementation
    :returns .abstracts.workflow.Workflow
    """

    def current_implementation(self):
        if hasattr(self, 'args'):
            return self.current()(self.args)
        else:
            return self.current()()

    def __get_versions(self):
        if not type(self.versions()) == list:
            raise ExternalError
        if not len(self.versions()) > 0:
            raise ExternalError
        for version in self.versions():
            if not issubclass(version, Workflow):
                raise ExternalError
        return self.versions()



import pytest


class CustomClass(object):
    def __init__(self, value):
        self.value = value


@pytest.mark.usefixtures("properties")
def test_from_args(properties):
    props = properties.from_(CustomClass(5))
    assert props == {'value': 5}


@pytest.mark.usefixtures("properties")
def test_from_args(properties):
    with pytest.raises(TypeError, match=r'Could not get properties from 5: .*'):
        properties.from_(5)
